public class OpenAddressingHashArray<K,D> {

   public static final byte LINEAR = 0;
   public static final byte QUADRATIC = 1;
   public static final byte DOUBLE_HASHING = 2;

   static final byte FREE = 0;
   static final byte OCCUPIED = 1;
   static final byte REMOVED = 2;

   K[] keys;
   D[] data;
   byte[] status;
   byte type;
   int hkprime;

   public OpenAddressingHashArray(int capacity, byte type) {
      this.type = type;
      if (type == QUADRATIC) {         
         capacity = adjustCapacityQuadratic(capacity);
      } else if (type == DOUBLE_HASHING) {
         capacity = adjustCapacityDouble(capacity);
      }
      this.keys = (K[]) new Object[capacity];
      this.data = (D[]) new Object[capacity];
      this.status = new byte[capacity];
   }

   int h(K key) {
      int hk = key.hashCode() % keys.length;
      return hk < 0 ? hk + keys.length : hk;
   }

   int s(K key, int i) {
      switch(this.type) {
      case LINEAR:
         return i;
      case QUADRATIC:
         if ((i & 1) == 0) {
            int ihalf = i >> 1;
            return (-ihalf * ihalf) % keys.length + keys.length; // ensure positive value
         } else {
            int ihalf = (i+1) >> 1;
            return ihalf * ihalf;
         }
      case DOUBLE_HASHING:
         if (i <= 1) {
            if (i == 0) {
               return 0; 
            } else {
               int hkp = key.hashCode() % (keys.length - 2);
               this.hkprime = hkp < 0 ? 1 + hkp + keys.length - 2 : 1 + hkp;
               return this.hkprime;
            }
         } else {
            return i * this.hkprime;
         }
      default:
         throw new RuntimeException("Unknown Hashtype");
      }
   }
   
   public D get(K key) {
      int hk = h(key);
      for (int i=0; i<keys.length; i++) {
         int pos = (hk + s(key, i)) % keys.length;
         switch (status[pos]) {
         case FREE:
            return null;
         case OCCUPIED:
            if (this.keys[pos].equals(key)) {
               return this.data[pos];
            }
         }
      }
      return null;
   }

   public void put(K key, D data) {
      int hk = h(key);
      for (int i=0; i<keys.length; i++) {
         int pos = (hk + s(key, i)) % keys.length;
         switch (status[pos]) {
         case REMOVED:
	     throw new RuntimeException("zu implementieren");
         case FREE:
            status[pos] = OCCUPIED;
            this.keys[pos] = key;
            this.data[pos] = data;
            return;
         case OCCUPIED:
            if (this.keys[pos].equals(key)) {
               this.data[pos] = data;
               return;
            }
         }
      }
      throw new RuntimeException("no space left");      
   }
   
   public void remove(K key) {
       throw new RuntimeException("zu implementieren");
   }
   
   public String toString() {
      StringBuilder b = new StringBuilder();
      for (int i=0; i < keys.length; i++) {
         if (i > 0) {
            b.append(", ");
         }
         b.append(i+"["+status[i]+"]");
         if (status[i] == OCCUPIED) {
            b.append(": " + keys[i]);
         }
      }
      return b.toString();
   }
   
   static int adjustCapacityQuadratic(int cap) {
      if (cap <= 3) {
         return 3;
      }
      int rem = (cap - 3) % 4;
      if (rem != 0) {
         cap += (4 - rem);
      }
      // now cap-3 is a multiple of 4
      while (!prime(cap)) {
         cap += 4;
      }
      return cap;
   }

   static int adjustCapacityDouble(int cap) {
      if (cap <= 3) {
         return 3;
      }
      cap |= 1;
      // now cap is an odd number >= 3
      while (!prime(cap)) {
         cap += 2;
      }
      return cap;
   }

   static boolean prime(int x) {
      if (x < 2) return false;
      int bound = (int)Math.sqrt(x) + 1;
      for (int i=2; i <= bound; i++) {
         if (x % i == 0) {
            return false;
         }
      }
      return true;
   }
}

